# Reaction Rate Field Plot: Reaction System Definition
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025
# Originally written in MATLAB in 2017

import pickle


def duplicate_figure(fig):
    pkl = pickle.dumps(fig)  # serialize figure
    fignew = pickle.loads(pkl)  # deserialize into fignew
    axnew = fignew.axes[0]
    return fignew, axnew



# Reaction Rate Field Plot: Reaction System Definition
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025
# Originally written in MATLAB in 2017

# Wishlist
# - Tab completion of reactor attributes and functions
# - Mark points and tau values on reactor curve:
# - - allow multiple points to be specified
# - Curves and tangents
# - - separate the individual paths of a CSTR curve
# - - draw CSTR tangents, especially when multiple tangents are possible

import numpy as np
import scipy as sci
import matplotlib.pyplot as plt


class ReactorRegistry:
    def __init__(self):
        self._store = {}

    def add(self, reactor):
        name = reactor.name
        if name in self._store:
            raise ValueError(f"Reactor {name!r} already registered.")
        self._store[name] = reactor
        setattr(self, name, reactor)

    def __getitem__(self, key):
        return self._store[key]

    def __iter__(self):
        return iter(self._store)  # iterates keys

    def __len__(self):
        return len(self._store)

    def items(self):
        return self._store.items()

    def keys(self):
        return self._store.keys()

    def values(self):
        return self._store.values()

    def __repr__(self):
        if not self._store:
            return "Reactors()"
        lines = ["Reactors("]
        for name, obj in self._store.items():
            lines.append(f"{name}:\n{obj!r}")
        lines.append(")")
        return "\n".join(lines)


reactors = ReactorRegistry()


class Reactor:
    """
    Container for reactor interacting with a reactive system.

    Reactor properties (name, flow, feed and time) must be specified as
    illustrated below.

    name:          name = 'PFR from fresh feed' (optional)
    type:          flow_type = 'pfr' (not case-sensitive)
                   PFR options:  'pfr', 'plug', 'plug flow', 'tube'
                   CSTR options: 'cstr', 'mixed', 'tank', 'stirred tank',
                                 'continuously stirred tank'
    feed:          feed = [1, 0] (specify according to axes)
    time constant: time = 10 (reactor is simulated from tau = 0 to this time)
    """

    def __init__(self, flow_type, feed, time=0, name='',
                 curve=None, tau=None, x=None, y=None, xy=None,
                 slope_x=None, slope_y=None):
        self.flow_type = flow_type
        self.feed = feed
        self.time = time
        self.name = name
        self.curve = []
        self.tau = np.array([])
        self.x = np.array([])
        self.y = np.array([])
        self.xy = np.array([])
        self.slope_x = []
        self.slope_y = []
        self.pfr_identifiers = (
            'pfr', 'plug', 'plug flow', 'tube'
        )
        self.cstr_identifiers = (
            'cstr', 'mixed', 'stirred tank', 'tank',
            'continuously stirred tank'
        )
        if reactors is not None:
            reactors.add(self)

    def __repr__(self):
        if np.array(self.tau).size == 0:
            disp_text = (f'Reactor:\n'
                     f'    name:\t{self.name}\n'
                     f'    flow type:\t{self.flow_type}\n'
                     f'    feed:\t{self.feed}')
        else:
            n_tau = np.array(self.tau).size
            disp_text = (f'Reactor:\n'
                     f'    name:\t{self.name}\n'
                     f'    flow type:\t{self.flow_type}\n'
                     f'    feed:\t{self.feed}\n'
                     f'    time:\t{self.time}\n'
                     f'    curve:\tarray of curve objects\n'
                     f'    tau:\t{n_tau}×1 numpy array of tau values\n'
                     f'    x:\t\t{n_tau}×1 numpy array of x values\n'
                     f'    y:\t\t{n_tau}×1 numpy array of y values\n'
                     f'    xy:\t\t{n_tau}×2 numpy array of [x, y] values\n'
                     f'    slope_i:\t{n_tau}×1 numpy arrays of slope data')
        return disp_text

    def simulate(self, system, ax, time_limit=10,
                 n_points=1000, show_feed=True):
        """
        Simulate and plot a reactor.

        reactor.simulate(time_limit=10, n_points=1000, show_feed=True)
            time_limit: default time limit for tau
            n_points:   number of integration or calculation points
            show_feed:  whether feed should be displayed as a point
        """
        def ratet(t, x):
            return system.rate(x)

        xF = self.feed
        if np.size(self.feed, 0) != 2:
            raise ValueError('The feed should have exactly 2 coordinates')
        if show_feed:
            ax.plot(xF[0], xF[1], 'ko', markersize=6)

        if self.flow_type.lower() in self.pfr_identifiers:
            if self.time == 0:
                int_t = time_limit
            else:
                int_t = self.time
            with np.errstate(all='ignore'):
                res = sci.integrate.solve_ivp(ratet, [0, int_t], xF,
                                              t_eval=np.linspace(
                                                0, int_t, n_points)
                                              )
                self.curve = ax.plot(res.y[0, :], res.y[1, :], 'r-')
                self.tau = np.array(res.t)
                self.x = np.array(res.y[0])
                self.y = np.array(res.y[1])
                self.xy = np.column_stack((self.x, self.y))
                rates_pfr = system.rate(res.y)
                self.slope_x = rates_pfr[0, :]
                self.slope_y = rates_pfr[1, :]

        if self.flow_type.lower() in self.cstr_identifiers:
            h_lim = system.h_lim
            v_lim = system.v_lim
            xx = np.linspace(h_lim[0], h_lim[1], n_points)
            yy = np.linspace(v_lim[0], v_lim[1], n_points)
            X, Y = np.meshgrid(xx, yy)

            def cstrpt(x, y):
                r = system.rate([x, y])
                with np.errstate(all='ignore'):
                    f = y - xF[1] - r[1] / r[0] * (x - xF[0])
                    return f

            self.curve = ax.contour(X, Y, cstrpt(X, Y),
                                          levels=[0],
                                          colors='b',
                                          linestyles='-',
                                          linewidths=1.5)
            xy = np.transpose(self.curve.allsegs[0][0])
            self.tau = np.array(
                (xy - np.array(xF)[:, None]) /
                system.rate([xy[0, :], xy[1, :]])
            )[0, :]
            # tau is calculated in duplicate from both axes, so use only one
            self.x = np.array(xy[0, :])
            self.y = np.array(xy[1, :])
            self.xy = np.column_stack((self.x, self.y))

            self.slope_x = self.x[1:] - self.x[:-1]
            self.slope_x = np.hstack(([self.slope_x[0]], self.slope_x[:-1]))
            self.slope_y = self.y[1:] - self.y[:-1]
            self.slope_y = np.hstack(([self.slope_y[0]], self.slope_y[:-1]))
            # Add dummy value to each slope array to make up for element
            # lost while taking difference
        return ax

    def plot_point(self, ax, tau=[], x=[], y=[],
                   near_point=[], distance=1.5):
        """
        Locate and plot a point on a reactor curve corresponding to a
        specified value of tau, x or y. Optionally, this point could be
        constrained to be within a specified distance from a specified point
        (near_point) in the concentration space.

        reactor.plot_point(ax, tau=0, x=0, y=0, near_point=[], distance=1.5)
            ax:         Axes on which the point should be plotted
            tau, x, y:  Scalars specifying the point to be locate At least one
                        of these inputs must be provided
                        If multiple inputs are present, they are matched
                        in the order tau, x, y
                        Thus, y will be matched only if tau and x are both []
                        And x will be matched only if tau is []
            near_point: 2x1 numpy array (defaults to reactor feed)
            distance:   Scalar indicating the search distance from the specified
                        point (defaults to 1.5)
                        If only one point meets the given criteria (tau,
                        x or y), set this distance to a large number,
                        e.g., 1.5
        """
        if not near_point:
            near_point = self.feed  # point defaults to reactor feed
        else:
            near_point = np.array(near_point)

        d = np.sqrt((self.x - near_point[0])**2 + (self.y - near_point[1])**2)
        xx = np.where(d < distance, self.x, np.nan)
        yy = np.where(d < distance, self.y, np.nan)
        tt = np.where(d < distance, self.tau, np.nan)

        # if statements are typed in reverse order of precedence
        if y and yy.any():
            idx = np.abs(yy - y).argmin()
        if x and xx.any():
            idx = np.abs(xx - x).argmin()
        if tau and tt.any():
            idx = np.abs(tt - tau).argmin()
        tau = tt[idx]
        x = xx[idx]
        y = yy[idx]
        label = f' $\\tau=$ {np.round(tau, 3)}, ({np.round(x, 3)}, {np.round(y, 3)})'
        if self.flow_type in self.pfr_identifiers:
            ax.plot(x, y, 'ro', markersize=6)
            ax.text(x, y, label, ha='left', va='center')
        elif self.flow_type in self.cstr_identifiers:
            ax.plot(x, y, 'bo', markersize=6)
            ax.text(x, y, label, ha='left', va='center')
        return ax

    def plot_tangent(self, ax, from_point, near_point=[], distance=1.5):
        """
        Plot a tangent to a reactor curve from a specified point
        (from_point). Optionally, the tangent could be constrained to
        touch the curve within a specified distance from another point
        (near_point) in the concentration space.

        reactor.plot_tangent(ax, from_point, near_point=[])
            ax:         Axes on which the tangent should be plotted
            from_point: Point from where tangent should be plotted, 2x1
                        numpy array
            near_point: 2x1 numpy array (defaults to reactor feed)
            distance:   Scalar indicating the search distance from the
                        specified point (defaults to 1.5)
                        If only one point meets the given criteria (tau,
                        x or y), set this distance to a large number,
                        e.g., 1.5
            tol:        Tolerance for tangent
                        Set to around 1e-2, adjust if no tangents are found
        """

        from_point = np.array(from_point)
        if not near_point:
            near_point = self.feed  # point defaults to reactor feed
        else:
            near_point = np.array(near_point)

        d = np.sqrt((self.x - near_point[0])**2 + (self.y - near_point[1])**2)
        xx = np.where(d < distance, self.x, np.nan)
        yy = np.where(d < distance, self.y, np.nan)

        xy = np.column_stack((xx, yy))

        angle_point_to_reactor = np.atan2(
            xy[:, 1] - from_point[1],
            xy[:, 0] - from_point[0]
        )

        angle_tangent = np.atan2(self.slope_y, self.slope_x)

        def angle_diff(a, b):
            return np.arctan2(np.sin(a - b), np.cos(a - b))

        idx = np.abs(
            angle_diff(angle_point_to_reactor, angle_tangent)
        ).argmin()

        if np.abs((angle_point_to_reactor[idx] - angle_tangent[idx])) < 1e-2:
            (line, ) = ax.plot((from_point[0], xy[idx, 0]),
                               (from_point[1], xy[idx, 1]), 'k-')
        else:
            print('Tangent not feasible from the given point (within tolerance)')
        return ax, line



# (C) Ganesh Sriram, 2025, gsriram@umd.edu
# Licensed under GNU Public License 3.0
# Originally written in MATLAB in 2018

# Wishlist
# - improve distillation design optimization
# - determine optimal feed tray
# - improve distillation block diagram and variable names in its function


class system:
    # Define a class 'system' to describe a ternary system
    
    def __init__(self, components, Antoine, Wilson_a, Wilson_b):

        self.components = components
        self.Antoine = Antoine
        self.Wilson_a = Wilson_a
        self.Wilson_b = Wilson_b

    # components: components in ternary mixture
    # Antoine: Antoine equation constants from NIST WebBook in NIST format
    # Wilson_a: a_ij parameters for Wilson activity coefficient correlation
    # Wilson_b: b_ij parameters for Wilson activity coefficient correlation
    # + diagonal elements a_ii and b_ii are always zero
    # + for an ideal liquid mixture, set all a_ij, b_ij to zero
    # + for non-ideal mixtures, obtain from literature or Aspen
    # + To get them from Aspen, create an Aspen file (apwz or apw) file
    #   containing the 3 compounds in the mixture
    # + Go to Methods on the left menu pane and select WILSON as the
    #   method with default settings
    # + Then navigate to Methods ≫ Parameters ≫ Binary Interaction ≫ Wilson
    # + Read the parameters a_ij and b_ij from the displayed table

    def __repr__(self):
        rows_Ant = '\n'.join(f'          {comp}: {row}' 
                           for comp, row in self.Antoine.items())
        rows_a = '\n'.join(f'        {comp}: {row}' 
                           for comp, row in self.Wilson_a.items())
        rows_b = '\n'.join(f'        {comp}: {row}' 
                           for comp, row in self.Wilson_b.items())
        disp_text = (f'System:\n'
                     f'    components: {self.components}\n'
                     f'    Antoine:\n{rows_Ant}\n'
                     f'    Wilson_b:\n{rows_a}\n'
                     f'    Wilson_b:\n{rows_b}\n')
        return disp_text


def make_system(components, Antoine, Wilson_a, Wilson_b):
    name = system(components=components,
                  Antoine=Antoine,
                  Wilson_a=Wilson_a,
                  Wilson_b=Wilson_b)
    return name


class properties:
    # Define a class 'properties' to describe a ternary mixture 
    # with a specific composition

    def __init__(self, P, molefrac, yeq, xeq, T_bubl, T_dew):
        self.P = P
        self.molefrac = molefrac
        self.yeq = yeq
        self.xeq = xeq
        self.T_bubl = T_bubl
        self.T_dew = T_dew

    def __repr__(self):
        x = self.molefrac
        disp_text = (f'Mixture VLE properties:\n'
                     f'    Pressure (P, bar): {self.P}\n'
                     f'    Overall composition (x): {x}\n'
                     f'    If this mixture were a saturated liquid:\n'
                     f'        Vapor in equilibrium (yeq): {self.yeq}\n'
                     f'        Bubble T (T_bubl, °C): {self.T_bubl}\n'
                     f'    If this mixture were a saturated vapor:\n'
                     f'        Liquid in equilibrium (xeq): {self.xeq}\n'
                     f'        Dew T (T_dew, °C): {self.T_dew} °C\n'
                     f'    Compositions are reported as mole fractions\n')
        return disp_text


class flash_res:
    # Define a class 'flash_res' to describe the results of a 
    # flash calculation

    def __init__(self, xF, x, y, T, P, Lfrac, Vfrac):
        self.xF = xF
        self.x = x
        self.y = y
        self.P = P
        self.T = T
        self.Lfrac = Lfrac
        self.Vfrac = Vfrac
        
    def __repr__(self):
        disp_text = (f'Flash results:\n'
                     f'    Feed: {self.xF}\n'
                     f'    Liquid product (x): {self.x}\n'
                     f'    Vapor product (y): {self.y}\n'
                     f'    Temperature (T, °C): {self.T}\n'
                     f'    Pressure (P, bar): {self.P}\n'
                     f'    Liquid fraction (Lfrac): {self.Lfrac}\n'
                     f'    Vapor fraction (Vfrac): {self.Vfrac}\n'
                     f'    Compositions are reported as mole fractions\n')
        return disp_text


class distill_res:
    # Define a class 'distill_res' to describe the results of a 
    # distillation calculation

    def __init__(self, xF, xD, xB, D, B, q, r, s, 
                 nstg, fstg, xLr, yVr, Tr, xLs, yVs, Ts, P):
        self.xF = xF
        self.xD = xD
        self.xB = xB
        self.D = D
        self.B = B
        self.q = q
        self.r = r
        self.s = s
        self.nstg = nstg
        self.fstg = fstg
        self.xLr = xLr
        self.yVr = yVr
        self.Tr = Tr
        self.xLs = xLs
        self.yVs = yVs
        self.Ts = Ts
        self.P = P
        
    def __repr__(self):
        stage_text_r = "Stagewise results (R-section):\n"
        stage_text_r += '         #'
        stage_text_r += '        xL (liq)                '
        stage_text_r += '        yV (vap)                '
        stage_text_r += '         T (°C)\n'
        
        for i in range(len(self.Tr)):
            xLr_row = '  '.join(f'{val:8.4f}' for val in self.xLr[i])
            yVr_row = '  '.join(f'{val:8.4f}' for val in self.yVr[i])
            stage_text_r += f'      {i+1:4d}      {xLr_row}    {yVr_row}    {self.Tr[i, 0]:8.2f}\n'
        
        stage_text_s = "Stagewise results (S-section):\n"
        stage_text_s += '         #'
        stage_text_s += '        xL (liq)                '
        stage_text_s += '        yV (vap)                '
        stage_text_s += '         T (°C)\n'

        for i in range(len(self.Ts)):
            xLs_row = '  '.join(f'{val:8.4f}' for val in self.xLs[i])
            yVs_row = '  '.join(f'{val:8.4f}' for val in self.yVs[i])
            stage_text_s += f'      {i+1:4d}      {xLs_row}    {yVs_row}    {self.Ts[i, 0]:8.2f}\n'

        
        disp_text = (f'Distillation parameters:\n'
                     f'    Feed: {self.xF}\n'
                     f'    Distillate composition (xD): {self.xD}\n'
                     f'    Bottoms composition (xB): {self.xB}\n'
                     f'    Distillate flow rate (D): {self.D}\n'
                     f'    Bottoms flow rate (B): {self.B}\n'
                     f'    Feed quality (q): {self.q}\n'
                     f'    Reflux ratio (r): {self.r}\n'
                     f'    Boilup ratio (s): {self.s}\n'
                     f'    Number of stages (nstg): {self.nstg}\n'
                     f'    Feed stage (fstg): {self.fstg}\n'
                     f'    Pressure (P, bar): {self.P}\n'
                     f'{stage_text_r}'
                     f'{stage_text_s}'
                     f'Compositions are reported as mole fractions\n'
                     f'Flows are molar relative to the feed\n')
        return disp_text


def vle_calc(system, axes, P=1.013,
             fsize=8, n_vectors=31, arrow_scale=31, xy_tol=-1e-6,
             Tmin=None, Tmax=None,
             contour_T_heavy=10, contour_T_medium=5, contour_T_light=1,
             plot_residue=False, residue_feeds=None, 
             integration_time=10, show_feeds=False):

    # Plot equilibrium vectors, constant-temperature contours and residue
    # curves for ternary vapor-liquid mixtures in [x_A,x_B] or [y_A,y_B] space
    # Perform flash and distillation calclations
    # A homogeneous non-ideal liquids phase is described by the Wilson
    # activity coefficient model; the vapor phase is assumed ideal
    # Vapor pressures are described by the Antoine equation in NIST format
    # (log10, bar, K)

    import numpy as np
    import scipy as sci
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    from collections import namedtuple
    import pickle
    import warnings
    
    # System properties

    components = system.components
    Antoine = system.Antoine
    Wilson_a = system.Wilson_a
    Wilson_b = system.Wilson_b

    assert axes is not None, 'Axes need to be specified'

    ax0 = False
    for j in components:  # check whether axis 0 is specified
        if j in axes:
            if axes[j] == 0:
                ax0 = True

    assert ax0, 'Axis 0 needs to be specified'

    ax1 = False
    for j in components:  # check whether axis 1 is specified
        if j in axes:
            if axes[j] == 1:
                ax1 = True

    assert ax1, 'Axis 1 needs to be specified'

    for j in components:  # identify implicit component if not specified
        if j not in axes:
            axes[j] = 2

    # Constants and functions

    T0 = 273.15  # 0 °C in K

    def Psat(T):  # vapor pressure (bar) as a function of T (°C)
        Psat = np.zeros(3)
        for j in components:
            assert 'A' in Antoine[j], \
                f'Antoine parameter A is missing for {j}'
            assert 'B' in Antoine[j], \
                f'Antoine parameter B is missing for {j}'
            assert 'C' in Antoine[j], \
                f'Antoine parameter C is missing for {j}'

            A = Antoine[j]['A']
            B = Antoine[j]['B']
            C = Antoine[j]['C']

            Ps = 10 ** (A - B / (T + T0 + C))
            Psat[axes[j]] = Ps
        return Psat

    if Tmin is None:  # minimum boiling point amongst the 3 components
        Tmin = np.inf
        for jj in range(3):
            res = sci.optimize.root_scalar(lambda T: (Psat(T) - P)[jj], x0=0)
            if res.root < Tmin:
                Tmin = res.root

    if Tmax is None:  # maximum boiling point amongst the 3 components
        Tmax = -T0
        for jj in range(3):
            res = sci.optimize.root_scalar(lambda T: (Psat(T) - P)[jj], x0=0)
            if res.root > Tmax:
                Tmax = res.root

    Tr = (Tmin + Tmax) / 2  # representative temperature of the system

    def gamma(x, T):
        # Wilson activity coefficient as a function of x and T (°C)

        A = np.zeros([3, 3])
        for i in components:
            for j in components:
                A[axes[i]][axes[j]] = np.exp(Wilson_a[i][j] +
                                             Wilson_b[i][j] /
                                             (T + T0))
        lng = 1 - np.log(A @ x) - np.transpose(A) @ (x / (A @ x))
        g = np.exp(lng)
        return g

    def delxt(t, x):
        # derivative of x with respect to t, expressed as f(t, x)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            res = sci.optimize.root_scalar(lambda T: 1 - np.sum(x *
                                                                gamma(x, T) *
                                                                Psat(T) / P),
                                           x0=Tr)
        T = res.root
        g = gamma(x, T)
        Ps = Psat(T)
        delx = x - x * g * Ps / P
        return delx

    def delx(x):
        # derivative of x with respect to t, expressed as f(x)

        res = sci.optimize.root_scalar(lambda T: 1 - np.sum(x *
                                                            gamma(x, T) *
                                                            Psat(T) / P),
                                       x0=Tr)
        T = res.root
        g = gamma(x, T)
        Ps = Psat(T)
        delx = x - x * g * Ps / P
        return delx, T

    def yeq(x):
        dx, T = delx(x)
        y = x - dx
        return y, T

    def xeq(y):
        Y = np.append(y, Tr)
        res = sci.optimize.root(lambda X: np.append(Y[0:3] * P - X[0:3] *
                                                    gamma(X[0:3], X[3]) *
                                                    Psat(X[3]),
                                                    np.sum(X[0:3]) - 1),
                                x0=Y)
        x = res.x[0:3]
        T = res.x[3]
        return x, T

    # Initialize figure

    fig, ax = plt.subplots(figsize=(fsize, fsize), facecolor='white')
    ax.set_aspect('equal', adjustable='box')
    for spine in ax.spines.values():
        spine.set_visible(True)
        spine.set_linewidth(1)
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    font = 12
    for j in components:
        if axes[j] == 0:
            ax.set_xlabel(j, fontsize=font)
        if axes[j] == 1:
            ax.set_ylabel(j, fontsize=font)
    ax.grid(True)

    # Calculate and plot equilibrium field

    X = np.linspace(0, 1, n_vectors)
    xx, yy = np.meshgrid(X, X)
    xx = np.where(1 - xx - yy < xy_tol, np.nan, xx)
    yy = np.where(1 - xx - yy < xy_tol, np.nan, yy)

    uu, vv = np.meshgrid(X, X)
    uu = np.where(1 - uu - vv < xy_tol, np.nan, uu)
    vv = np.where(1 - uu - vv < xy_tol, np.nan, vv)

    for i in range(np.size(xx, 0)):
        for j in range(np.size(yy, 1)):
            if ~np.isnan(xx[i, j]) and ~np.isnan(yy[i, j]):
                uv, _ = delx([xx[i, j], yy[i, j], 1 - xx[i, j] - yy[i, j]])
                uu[i, j], vv[i, j] = uv[0], uv[1]

    mag = np.hypot(uu, vv)

    un, vn = uu, vv

    with np.errstate(invalid='ignore'):
        un = np.where(np.isnan(un), np.nan, np.where(mag > 0,
                                                     np.divide(uu, mag), 0))
        vn = np.where(np.isnan(vn), np.nan, np.where(mag > 0,
                                                     np.divide(vv, mag), 0))

    ax.quiver(X, X, un, vn, color='#8888ff', scale=arrow_scale, alpha=0.4)

    # Plot residue curve if requested

    if plot_residue:
        xF = np.zeros(3)  # initialize feed vector
        assert residue_feeds is not None, \
            'Feed point(s) must be specified if residue curve is requested'

        row_sums = {i: sum(d.values()) for i, d in residue_feeds.items()}
        for i, total in row_sums.items():
            assert abs(total - 1) < 1e-8, (
                'Residue curve feed mole fractions must add up to 1\n' +
                f'Row {i} does not sum to 1 (got {total})')
        
        for i in residue_feeds:
            for j in residue_feeds[i]:
                xF[axes[j]] = residue_feeds[i][j]
            if show_feeds:
                ax.plot(xF[0], xF[1], 'ko', markersize=6)
            with np.errstate(all='ignore'):
                int_t = integration_time
                res = sci.integrate.solve_ivp(delxt, [0, int_t], xF,
                                              t_eval=np.linspace(0, int_t, 100))
                ax.plot(res.y[0, :], res.y[1, :], 'k-')
                res = sci.integrate.solve_ivp(delxt, [0, -int_t], xF,
                                              t_eval=np.linspace(0, -int_t, 100))
                ax.plot(res.y[0, :], res.y[1, :], 'k-')

    # Calculate and plot temperature contours

    x_contours = 51  # spacing of x for contour plotting
    XT = np.linspace(0, 1, x_contours)
    X0, X1 = np.meshgrid(XT, XT)
    Y0, Y1 = np.meshgrid(XT, XT)
    TT0, TT1 = np.meshgrid(XT, XT)

    for i in range(np.size(XT)):
        for j in range(np.size(XT)):
            if XT[i] + XT[j] <= 1:
                x = [XT[i], XT[j], 1 - XT[i] - XT[j]]
                res = sci.optimize.root_scalar(lambda T:
                                               1 - np.sum(x * gamma(x, T) *
                                                          Psat(T) / P),
                                               x0=Tr)
                T = res.root

                X0[i, j] = XT[i]
                X1[i, j] = XT[j]
                TT0[i, j] = T

                y = x * gamma(x, T) * Psat(T) / P
                Y0[i, j] = y[0]
                Y1[i, j] = y[1]
                TT1[i, j] = T
            else:
                TT0[i, j] = np.nan
                TT1[i, j] = np.nan

    def clevels(c):  # contour levels
        tmin = np.round(Tmin / c) * c
        tmax = np.round(Tmax / c + 1) * c
        clevels = np.arange(tmin, tmax, c)
        return clevels

    cl = ax.contour(X0, X1, TT0,
                    levels=clevels(contour_T_heavy),
                    linewidths=0.5, colors='#0000ff')
    ax.clabel(cl, inline=True, fontsize=8)
    ax.contour(X0, X1, TT0,
               levels=clevels(contour_T_medium),
               linewidths=0.25, colors='#0000ff')
    ax.contour(X0, X1, TT0,
               levels=clevels(contour_T_light),
               linewidths=0.05, colors='#0000ff')

    cl = ax.contour(Y0, Y1, TT1,
                    levels=clevels(contour_T_heavy),
                    linewidths=0.5, colors='#ff0000')
    ax.clabel(cl, inline=True, fontsize=8)
    ax.contour(Y0, Y1, TT1,
               levels=clevels(contour_T_medium),
               linewidths=0.25, colors='#ff0000')
    ax.contour(Y0, Y1, TT1,
               levels=clevels(contour_T_light),
               linewidths=0.05, colors='#ff0000')

    # Wrap up, save and display plot

    ax.plot([0, 1], [0, 0], 'k-')  # right triangle, horizontal line
    ax.plot([0, 0], [0, 1], 'k-')  # right triangle, vertical line
    ax.plot([0, 1], [1, 0], 'k-')  # right triangle, diagonal line

    plt.draw()

    def mix_props(x):
        ye, Tb = yeq(x)
        xe, Td = xeq(x)
        p = properties(P=P,
                       molefrac=x,
                       yeq=ye,
                       xeq=xe,
                       T_bubl=Tb,
                       T_dew=Td)
        return p

    def atanm(x):
        # Modified arctan function, returns a result in the range (0, {pi})
        res = np.atan(x)
        if res < 0:
            res = res + np.pi
        return res

    def angle_dx(x):
        dx, T = delx(x)
        if abs(dx[0]) <= 1e-6:
            angle = np.pi / 2
        else:
            slope = dx[1] / dx[0]
            angle = atanm(slope)
        return angle

    def angle_xy(x, y):
        if np.abs(x[0] - y[0]) <= 1e-6:
            angle = np.pi / 2
        else:
            slope = (x[1] - y[1]) / (x[0] - y[0])
            angle = atanm(slope)
        return angle

    def flashT_obj(x, xF, T):
        z = ((angle_dx(x) - angle_xy(x, xF)) ** 2 +         
             (yeq(x)[1] - T) ** 2 +
             (xeq(yeq(x)[0])[1] - T) ** 2)
        return z
   
    def flashT(xF, T):
        eq_cons = {'type': 'eq',
               'fun' : lambda x: np.sum(x) - 1}
    
        bounds = sci.optimize.Bounds([0, 0, 0], [1, 1, 1])
        
        res = sci.optimize.minimize(lambda x: flashT_obj(x, xF, T),
                                    x0=xF,
                                    method='SLSQP',
                                    constraints=[eq_cons],
                                    bounds=bounds,
                                    tol=1e-9)
        
        x = res.x
        y = yeq(x)[0]
        T = yeq(x)[1]
        Vfrac = np.linalg.norm(x - xF) / np.linalg.norm(x - y)
        
        res = flash_res(xF=xF,
                        x=x,
                        y=y,
                        T=T,
                        P=P,
                        Lfrac=1 - Vfrac,
                        Vfrac=Vfrac)
        return res

    def flashVfrac_obj(x, xF, Vfrac):
        x = np.array(x)
        Tx = mix_props(x).T_bubl
        xF = np.array(xF)
        y, Ty = yeq(x)
        
        z = ((angle_dx(x) - angle_xy(x, xF)) ** 2 +         
             (np.linalg.norm(x - xF) / np.linalg.norm(x - y) - Vfrac) ** 2 +
             (Tx - Ty) ** 2)
        return z
        
    def flashVfrac(xF, Vfrac):
        eq_cons = {'type': 'eq',
               'fun' : lambda x: np.sum(x) - 1}
    
        bounds = sci.optimize.Bounds([0, 0, 0], [1, 1, 1])
        
        res = sci.optimize.minimize(lambda x: flashVfrac_obj(x, xF, Vfrac),
                                    x0=xF,
                                    method='SLSQP',
                                    constraints=[eq_cons],
                                    bounds=bounds,
                                    tol=1e-9)

        x = res.x
        y = yeq(x)[0]
        T = yeq(x)[1]
        Vfrac = np.linalg.norm(x - xF) / np.linalg.norm(x - y)
        
        res = flash_res(xF=xF,
                        x=x,
                        y=y,
                        T=T,
                        P=P,
                        Lfrac=1 - Vfrac,
                        Vfrac=Vfrac)
        return res
    
    def flash_bfd(flash_res):
        # Draw a BFD of a flash evaporator

        res = flash_res

        P_str = np.round(res.P, 2)
        T_str = np.round(res.T, 2)
        xF_str = np.round(res.xF, 2)

        if res.x is None:
            x_str = '?'
        else:
            x_str = np.round(res.x, 2)

        if res.y is None:
            y_str = '?'
        else:
            y_str = np.round(res.y, 2)

        if res.Lfrac is None:
            Lfrac_str = '?'
        else:
            Lfrac_str = np.round(res.Lfrac, 2)

        if res.Vfrac is None:
            Vfrac_str = '?'
        else:
            Vfrac_str = np.round(res.Vfrac, 2)

        fig1, ax1 = plt.subplots()
        plt.axis('off')
        _ = ax1.set_xlim(-4, 4)
        _ = ax1.set_ylim(-4, 4)

        patches.ArrowStyle.Curve

        feed = patches.Arrow(-3, 0, 2, 0, width=0.5, color='black')
        ax1.add_patch(feed)

        ax1.text(-1.1, 0.2,
                f'xF = {xF_str}',
                horizontalalignment='right',
                verticalalignment='bottom',
                fontsize=8,
                color='black')

        vap = patches.Arrow(1, 1.6, 2, 0, width=0.5, color='red')
        ax1.add_patch(vap)

        ax1.text(1.1, 1.8,
            f'y = {y_str}',
            horizontalalignment='left',
            verticalalignment='bottom',
            fontsize=8,
            color='red')

        liq = patches.Arrow(1, -1.6, 2, 0, width=0.5, color='blue')
        ax1.add_patch(liq)

        ax1.text(1.1, -1.4,
            f'x = {x_str}',
            horizontalalignment='left',
            verticalalignment='bottom',
            fontsize=8,
            color='blue')

        unit = patches.Rectangle((-1, -2), 2, 4,
                                 linewidth=2,
                                 edgecolor='black',
                                 facecolor='white',
                                 alpha=1)
        ax1.add_patch(unit)

        ax1.text(0, 0.5, 'Flash',
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=10)

        ax1.text(0, 0, f'P = {P_str} bar',
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=8)

        ax1.text(0, -0.5, f'T = {T_str} °C',
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=8)

        ax1.text(2, 0, f'Vfrac = {Vfrac_str}',
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=8)

        return fig1, ax1

    def distill(xF, D=None, xD=None, B=None, xB=None,
                r=None, s=None, q=None,
                nstg=None, fstg=None,
                mode=None, calc_begin=None,
                plot_distill=True):

        # nstg and fstg are 1-based (provided by user)
        # Nstg and Fstg are 0-based (used in this function)

        if nstg is not None:
            Nstg = nstg - 1
        if fstg is not None:
            Fstg = fstg - 1
        
        if mode in ('minstages', 'minreflux'):
            assert xD is not None or xB is not None, (
                "Either xD or xB must be specified " +
                "if mode is 'minstages' or 'minreflux'")

            if xD is not None:
                assert D is not None, (
                    'If xD is specified, D must be specified')

            if xB is not None:
                assert B is not None, (
                    'If xB is specified, B must be specified')

            if xD is None:
                D = 1 - B
                xD = (xF - B * np.array(xB)) / D
            elif xB is None:
                B = 1 - D
                xB = (xF - D * np.array(xD)) / B

            xD = np.array(xD)
            xB = np.array(xB)

            s = (
                q - 1 
                + (q + r) * np.linalg.norm(xF - xB)
                  / np.linalg.norm(xD - xF)
            )
            
        xF = np.array(xF)

        assert mode in (None, 'minstages', 'minreflux'), (
            f"mode must be set to None (default calculation), "
            f"'minstages' or 'minreflux' (got {mode})")

        count_qrs = sum(x is not None for x in (q, r, s))

        if mode in (None, 'minreflux'):
            assert count_qrs >= 2, (
                'feed quality (q) and reflux ratio (r) must be specified')

        if mode is None:
            assert nstg is not None, (
                'Number of stages (nstg) must be specified')
            assert fstg is not None, (
                'Feed stage (fstg) must be specified')    
            stages_r = np.arange(0, Fstg)
            stages_s = np.arange(Nstg, Fstg, -1)
        elif mode == 'minstages':
            r = 1e+6                      
            assert nstg is not None, (
                'Number of stages (nstg) must be specified')  
            if calc_begin is None:
                calc_begin = 'distillate'
            assert calc_begin in ('distillate', 'bottoms'), (
                f"For 'minstages' mode, calc_begin must either be " +
                f"'distillate' or 'bottoms' (got {calc_begin})")
            if calc_begin == 'distillate':
                stages_r = np.arange(0, Nstg)
            elif calc_begin == 'bottoms':
                stages_s = np.arange(Nstg, 0, -1)
        elif mode == 'minreflux':
            nstg = 200
            Nstg = nstg - 1
            stages_r = np.arange(0, Nstg - 1)
            stages_s = np.arange(Nstg, 1, -1)

        xLr = np.full((nstg, 3), np.nan)
        yVr = np.full((nstg, 3), np.nan)
        Tr = np.full((nstg, 1), np.nan)
        
        xLs = np.full((nstg, 3), np.nan)
        yVs = np.full((nstg, 3), np.nan)
        Ts = np.full((nstg, 1), np.nan)

        if mode in ('minstages', 'minreflux'):
            # Rectifying section profile
            
            xLr[0, :] = xD
            yVr[0, :] = yeq(xD)[0]
            Tr[0] = yeq(xD)[1]
            
            if ((mode in (None, 'minreflux')) or
                (mode == 'minstages' and calc_begin == 'distillate')):        
                for j in stages_r:
                    yVr[j + 1, :] = r / (r + 1) * xLr[j, :] + 1 / (r + 1) * np.array(xD)
                    xLr[j + 1, :] = xeq(yVr[j + 1, :])[0]
                    Tr[j + 1] = xeq(yVr[j + 1, :])[1]
                
            # Stripping section profile
            
            xLs[nstg-1, :] = xB
            yVs[nstg-1, :] = yeq(xB)[0]
            Ts[nstg-1] = yeq(xB)[1]
    
            if ((mode in (None, 'minreflux')) or 
                (mode == 'minstages' and calc_begin == 'bottoms')):     
                for j in stages_s:
                    xLs[j - 1, :] = s / (s + 1) * yVs[j, :] + 1 / (s + 1) * np.array(xB)
                    yVs[j - 1, :] = yeq(xLs[j - 1, :])[0]
                    Ts[j - 1] = yeq(xLs[j - 1, :])[1]
        else:
            def distill_obj(xD):
                # D is input by user, xD is the optimization variable

                B = 1 - D
                xB = (xF - D * np.array(xD)) / B

                assert 0 <= q <= 1, 'Provide q in the range (0, 1)'

                s = (
                    q - 1 
                    + (q + r) * np.linalg.norm(xF - xB)
                      / np.linalg.norm(xD - xF)
                )
                
                res = vle3.flashVfrac(xF, q)
                xF_obj = res.x
                yF_obj = res.y

                obj = 0

                # Rectifying section profile
            
                xLr[0, :] = xD
                yVr[0, :] = yeq(xD)[0]
                Tr[0] = yeq(xD)[1]
                
                for j in stages_r:
                    yVr[j + 1, :] = r / (r + 1) * xLr[j, :] + 1 / (r + 1) * np.array(xD)
                    xLr[j + 1, :] = xeq(yVr[j + 1, :])[0]
                    Tr[j + 1] = xeq(yVr[j + 1, :])[1]

                obj += (
                    np.linalg.norm(xF_obj - xLr[Fstg]) ** 2 +
                    np.linalg.norm(yF_obj - yVr[Fstg]) ** 2
                )
                
                # Stripping section profile
                
                xLs[nstg-1, :] = xB
                yVs[nstg-1, :] = yeq(xB)[0]
                Ts[nstg-1] = yeq(xB)[1]
        
                for j in stages_s:
                    xLs[j - 1, :] = s / (s + 1) * yVs[j, :] + 1 / (s + 1) * np.array(xB)
                    yVs[j - 1, :] = yeq(xLs[j - 1, :])[0]
                    Ts[j - 1] = yeq(xLs[j - 1, :])[1]

                obj += (
                    np.linalg.norm(xF_obj - xLs[Fstg]) ** 2 +
                    np.linalg.norm(yF_obj - yVs[Fstg]) ** 2
                )
                
                return obj
            
            # print(distill_obj(xF))
            # print(distill_obj(np.array([1, 0, 0])))
            # print(distill_obj(np.array([0, 0, 1])))
            
            eq_cons = {'type': 'eq',
               'fun' : lambda x: np.sum(x) - 1}

            bounds = sci.optimize.Bounds([0, 0, 0], [1, 1, 1])

            res = sci.optimize.minimize(lambda x: distill_obj(x),
                                        x0=xF+0.01,
                                        method='SLSQP',
                                        constraints=[eq_cons],
                                        bounds=bounds,
                                        tol=1e-3)

            # print(res.fun)
            # print(res)
            
            xD = res.x

            B = 1 - D
            xB = (xF - D * np.array(xD)) / B

            s = (
                q - 1 
                + (q + r) * np.linalg.norm(xF - xB)
                  / np.linalg.norm(xD - xF)
            )

            # Rectifying section profile
        
            xLr[0, :] = xD
            yVr[0, :] = yeq(xD)[0]
            Tr[0] = yeq(xD)[1]
            
            for j in stages_r:
                yVr[j + 1, :] = r / (r + 1) * xLr[j, :] + 1 / (r + 1) * np.array(xD)
                xLr[j + 1, :] = xeq(yVr[j + 1, :])[0]
                Tr[j + 1] = xeq(yVr[j + 1, :])[1]
                
            # Stripping section profile
            
            xLs[nstg-1, :] = xB
            yVs[nstg-1, :] = yeq(xB)[0]
            Ts[nstg-1] = yeq(xB)[1]
    
            for j in stages_s:
                xLs[j - 1, :] = s / (s + 1) * yVs[j, :] + 1 / (s + 1) * np.array(xB)
                yVs[j - 1, :] = yeq(xLs[j - 1, :])[0]
                Ts[j - 1] = yeq(xLs[j - 1, :])[1]
    
        res = distill_res(xF=xF, xB=xB, xD=xD, D=D, B=B, q=q, r=r, s=s,
                          nstg=nstg, fstg=fstg,
                          xLr=xLr, yVr=yVr, Tr=Tr,
                          xLs=xLs, yVs=yVs, Ts=Ts,
                          P=P)
        
        return res

    def distill_bfd(distill_res):
        # Draw a BFD of a distillation column

        res = distill_res

        P_str = np.round(res.P, 2)
        xF_str = np.round(res.xF, 2)

        if res.xB is None:
            x_str = '?'
        else:
            x_str = np.round(res.xB, 2)

        if res.xD is None:
            y_str = '?'
        else:
            y_str = np.round(res.xD, 2)

        fig1, ax1 = plt.subplots()
        plt.axis('off')
        _ = ax1.set_xlim(-4, 4)
        _ = ax1.set_ylim(-4, 4)

        patches.ArrowStyle.Curve

        feed = patches.Arrow(-3, 0, 2, 0, width=0.5, color='black')
        ax1.add_patch(feed)

        ax1.text(-1.1, 0.2,
                f'xF = {xF_str}',
                horizontalalignment='right',
                verticalalignment='bottom',
                fontsize=8,
                color='black')

        vap = patches.Arrow(1, 2.6, 2, 0, width=0.5, color='red')
        ax1.add_patch(vap)

        ax1.text(1.1, 2.8,
            f'xD = {y_str}',
            horizontalalignment='left',
            verticalalignment='bottom',
            fontsize=8,
            color='red')

        liq = patches.Arrow(1, -2.6, 2, 0, width=0.5, color='blue')
        ax1.add_patch(liq)

        ax1.text(1.1, -2.4,
            f'xB = {x_str}',
            horizontalalignment='left',
            verticalalignment='bottom',
            fontsize=8,
            color='blue')

        unit = patches.Rectangle((-1, -3), 2, 6,
                                 linewidth=2,
                                 edgecolor='black',
                                 facecolor='white',
                                 alpha=1)
        ax1.add_patch(unit)

        ax1.text(0, 0.5, 'Column',
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=10)

        ax1.text(0, 0, f'P = {P_str} bar',
                 horizontalalignment='center',
                 verticalalignment='center',
                 fontsize=8)
        
        return fig1, ax1

    def duplicate(fig):
        pkl = pickle.dumps(fig)  # serialize figure
        fignew = pickle.loads(pkl)  # deserialize into fignew
        axnew = fignew.axes[0]
        
        return fignew, axnew
    
    bundle = namedtuple('vle3', ['Psat', 'yeq', 'xeq',
                                 'mix_props', 'flashT', 'flashVfrac',
                                 'distill', 'flash_bfd', 'distill_bfd',
                                 'fig', 'ax', 'plt', 'duplicate'])
    
    return bundle(Psat=Psat, yeq=yeq, xeq=xeq,
                  mix_props=mix_props, flashT=flashT,
                  flashVfrac=flashVfrac, distill=distill,
                  flash_bfd=flash_bfd, distill_bfd=distill_bfd,
                  fig=fig, ax=ax, plt=plt, duplicate=duplicate)



def simplex_tableaux(c,A_ub=[],b_ub=[],var_names=[],slack_names=[],tol=1e-4,dec=2,colormap='Greys'):
    # (C) Ganesh Sriram, gsriram@umd.edu
    # Licensed under GNU Public License 3.0
    
    # This script implements the simplex method for solving LPs in the form of matrix operations
    # It was written specifically for CHBE444: ChE Design I, University of Maryland
  
    # This version supports only inequality constraints (equality constraints are not permitted)
    # This version does not update the auxiliary after each pivot

    # Changelog
    # v. 1.3
    #     A and b are now called A_ub and b_ub for consistency with linprog
    #     var_names and/or slack_names may be left blank if user wants the algorithm to create them
    #     tol is used to avoid lots of iterations with little/no incremental benefit
    #     Tlist is returned as a list of pandas dataframes with headers
     
    # c: cost vector for objective function
    # A_ub: inequality matrix (coefficients of x_i in inequalities)
    # b_ub: inequality vector (constants in inequalities)
    # var_names and slack_names: headers
    # A and c should have the same number of columns (equal to the number of variables)
    # A and b should have the same number of rows (equal to the number of constraints)
    
    
    import numpy as np
    import pandas as pd


    A = A_ub
    b = b_ub
    
    # Examine vectors and matrix dimensions for consistency
    # Create headers (variable and slack names) if not provided
    
    Arows = np.size(A,axis=0)
    Acols = np.size(A,axis=1)
    brows = np.size(b)
    ccols = np.size(c)

    assert Acols == ccols, 'Check input dimensions. A and c have different numbers of columns.'
    assert Arows == brows, 'Check input dimensions. A and b have different numbers of rows.'

    if var_names == []:  # create var_names if empty
        var_names = [f'x{j}' for j in range(1,1+Acols)]  # Acols and ccols would have been checked for consistency by now                                                      
    
    var_num = np.size(var_names)
    var_max = max(int(s.lstrip('x')) for s in var_names)

    if slack_names == []:
        slack_names = [f'x{i}' for i in range(var_max+1,var_max+1+Arows)]  # Arows and brows would have been checked for consistency by now
    
    slack_min = min(int(s.lstrip('x')) for s in slack_names)
    
    if slack_names == [] or slack_min <= var_max:  # create slack_names if empty or index is <= var_max
        slack_names = [f'x{i}' for i in range(var_max+1,var_max+1+Arows)]  # Arows and brows would have been checked for consistency by now
            
    slack_num = np.size(slack_names)
        
    assert Acols == var_num, 'Check input dimensions. The number of original variables is different from the number of columns of A.'    
    assert Arows == slack_num, 'Check input dimensions. The number of slack variables is different from the number of rows of A.'
       
    
    # Process LP and create initial tableau
    
    n = np.size(c)  # count number of original variables
    m = np.size(b)  # count number of slack variables
    
    def sleq(i,m):  # function to get the ith row of a mxm identity matrix
        I = np.identity(m)
        return I[i]

    T = np.concatenate(([1],c,np.zeros(m),[0]),0)  # construct initial tableau T
   
    for i in range(0,m):
        t = np.concatenate(([0],A[i],sleq(i,m),[b[i]]),0)  # current row
        T = np.vstack((T,t))  # append current row below T
    
    Trows = np.size(T,axis=0)
    Tcols = np.size(T,axis=1)
    
    aux_needed = False  # compose auxiliary objective if needed
    
    c_aux = np.zeros(Tcols)  # initialize cost vector for auxiliary objective
    for i in range(1,1+m):
        if T[i,Tcols-1] < 0:
            aux_needed = True
            c_aux += T[i]  # update cost vector for auxiliary objective
    
    if aux_needed == True:
        T = np.insert(T,1,c_aux,axis=0)
        t = np.concatenate(([0],[1],np.zeros(m)),0)
        T = np.insert(T,1,t,axis=1)
    
    if aux_needed == False:
        row_names = ["z"]
    else:
        row_names = ["z","z'"]
       
    for k in range(0,np.size(slack_names)):
        row_names.append(slack_names[k])
    
    if aux_needed == False:
        col_names = ["z"]
    else:
        col_names = ["z","z'"]
    
    for j in range(0,np.size(var_names)):
        col_names.append(var_names[j])
    for i in range(0,np.size(slack_names)):
        col_names.append(slack_names[i])
    col_names.append('b')

    Tlist = [pd.DataFrame(T,columns=col_names,index=row_names)]  # store tableau in a list
    
    
    # Use pandas to display T as a tableau
    
    def conditionally_format_initial(styler):
        styler.background_gradient(axis=None,vmin=np.min(T),vmax=np.max(T),cmap=colormap)
        return styler
   
    df = pd.DataFrame(T,columns=col_names,index=row_names)
    df_style = df.style.format(precision=dec).pipe(conditionally_format_initial).set_properties(
        **{'text-align':'center'})
    df_style.set_table_styles([{'selector':'th.col_heading','props':'text-align: center;'},],overwrite=False)
    display(df_style)  
    
    Trows = np.size(T,axis=0)
    Tcols = np.size(T,axis=1)

    if aux_needed == True:
        start = 2
        index = 1
    else:
        start = 1
        index = 0

        
    # Iterate
    
    stop_condition = False
    
    while stop_condition == False:
        if aux_needed == True and T[1,Tcols-1] >= 0:
            aux_needed = False
            index = 0
        
        pcol = 0  # initialize pivot column
        cval = tol 
            
        for j in range(start,start+m+n):
            if T[index,j] < tol and T[index,j] < cval:
                cval = T[index,j]
                pcol = j  # identify pivot column
        if cval == 0:
            stop_condition = True
       
        if stop_condition == False:
            prow = 0  # initialize pivot row
            rval = np.inf
            for i in range(start,start+m):
                if T[i,pcol] > 0 and T[i,Tcols-1]/T[i,pcol] < rval:
                    rval = T[i,Tcols-1]/T[i,pcol]
                    prow = i  # identify pivot row

                       
            # Perform pivot
            
            R = np.identity(Trows)  # initialize pivoting matrix R to identity matrix
            
            factors = -T[:,pcol]/T[prow,pcol]  # factors to ensure that pivot column elements above and below pivot 
                                               # row are zero after pivot
            factors[prow] = 1;  # pivot row is unaltered during pivot
            R[:,prow] = factors
            
            T = np.dot(R,T)  # pivot operation
            
            row_names[prow] = col_names[pcol]  # update row headers

            Tlist.append(pd.DataFrame(T,columns=col_names,index=row_names))  # append tableau to the list
            
            
            def conditionally_format_current(styler):
                styler.background_gradient(axis=None,vmin=np.min(T),vmax=np.max(T),cmap=colormap)
                return styler
            
            df = pd.DataFrame(T,columns=col_names,index=row_names)
            df_style = df.style.format(precision=dec).pipe(conditionally_format_current).set_properties(
                **{'text-align':'center'})
            df_style.set_table_styles([{'selector':'th.col_heading','props':'text-align: center;'},],overwrite=False)
            display(df_style)
               
    return Tlist  # return the list of tableaux as dataframes


# Reaction Rate Field Plot: Reaction System Definition
# Written for CHBE444: Design I, University of Maryland
# (C) Ganesh Sriram, 2025
# Originally written in MATLAB in 2017

# Wishlist
# - Inequality constraints on axis variables to account for mass balances
# - Allow rates to be specified in terms of partial pressures
# - Allow units other than SI or METCBAR

import numpy as np
import matplotlib.pyplot as plt


class ReactionSystem:
    """
    Container for components and reactions in a reactive system.

    Please specify system properties (attributes) as illustrated
    below.

    Components:
        component_ids = ('A', 'B', 'C')
        component_names = {'A': 'n-pentane',
                           'B': '2-methylbutane',
                           'C': '2,2-dimethylpropane'}
    Use component_id for short names (EtOH) or molecular formulas
    Use component_names for full names

    Axes:
        axes = {0: 'A', 1: 'B'}

    Reaction indices:
        reactions = {'r1', 'r2'}

    Stoichiometry
        stoich = {'r1': {'A': -1, 'B': 1},
                  'r2': {'B': -1, 'C': 1}}

    Short-form kinetics
        kinetics = {'r1': lambda C: 3 * x['A'],
                    'r2': lambda C: 2 * x['B']}

    Long-form kinetics
        k1 = 3  # s^-1
        k2 = 2  # s^-1
        def r1(C):
            return k1 * x['A']
        def r2(C):
            return k2 * x['B']
        kinetics = {'r1': r1,
                    'r2': r2}

    Note on kinetics
        x['A'], x['B'], etc., are concentration (or equivalent) variables
        along the axes corresponding to 'A', 'B', etc. In some problems,
        it may be convenient to use moles or partial pressures for x[i].
    """

    def __init__(self, component_ids, component_names, axes, h_lim, v_lim,
                 reactions, stoich, kinetics):
        self.component_ids = tuple(component_ids)
        self.components = tuple(self.component_ids)
        self.component_names = component_names
        self.names = self.component_names
        self.axes = axes
        self.h_lim = h_lim
        self.v_lim = v_lim
        self.reactions = reactions
        self.stoich = stoich
        self.kinetics = kinetics

    def __repr__(self):
        rows_stoich = '\n'.join(f'        {rxn}: {s}'
                                for rxn, s in self.stoich.items())
        rows_kinetics = '\n'.join(f'        {rxn}: {k}'
                                  for rxn, k in self.kinetics.items())
        disp_text = (f'Reaction system:\n'
                     f'    component ids:\n\t{self.components}\n'
                     f'    component names:\n\t{self.component_names}\n'
                     f'    axes:\n\t{self.axes}\n'
                     f'    horizontal axis limits:\n\t{self.h_lim}\n'
                     f'    vertical axis limits:\n\t{self.v_lim}\n'
                     f'    reactions:\n\t{sorted(self.reactions)}\n'
                     f'    stoichiometry:\n{rows_stoich}\n'
                     f'    kinetics:\n{rows_kinetics}\n')
        return disp_text

    def rate(self, x):
        """ Reaction rate as a function of concentration vector x."""

        X = {'A': 0, 'B': 0}  # X is the concentration vector (or equivalent)
        r_rxn = dict.fromkeys(self.reactions, 0)
        r_comp = dict.fromkeys(self.axes, 0)

        for i in self.components:
            if i in self.axes:
                X[i] = x[self.axes[i]]

        for j in self.reactions:
            r_rxn[j] += self.kinetics[j](X)

        for i in self.components:
            if i in self.axes:
                for j in self.reactions:
                    if i in self.stoich[j]:
                        r_comp[i] += self.stoich[j][i] * r_rxn[j]

        r_comp = np.array(list(r_comp.values()))
        return r_comp

    def plot_rate_field(
        self,
        n_vec=51, n_vec_h=None, n_vec_v=None,
        arrow_scale=None, arrow_scale_h=None, arrow_scale_v=None,
        fsize=12, font=12
        ):

        """
        Plot rate field in concentration space.

        Example usage:
            system.draw_rate_field(
                n_vec_h=51, n_vec_v=51,
                fsize=12, font=12)

            n_vec: number of rate vectors along any dimension, defaults to 51
            n_vec_h: number of rate vectors along horizontal dimension (x-axis)
                     defaults to 51, can also be set by n_vec
            n_vec_v: number of rate vectors along vertical dimension (x-axis)
                     defaults to 51, can also be set by n_vec
            arrow_scale: scale of arrow along any dimension, defaults to n_vec
            arrow_scale_h: scale of arrow along horizontal dimension,
                           defaults to n_vec_x
            arrow_scale_v: scale of arrow along vertical dimension,
                           defaults to n_vec_y
            fsize: figure size, defaults to 12
            font: main font used for axis label, etc., defaults to 12
            h_lim, v_lim: horizontal and vertical axis limits,
            default to [0, 1]
        """

        fig, ax = plt.subplots(figsize=(fsize, fsize), facecolor='white')
        h_lim = self.h_lim
        v_lim = self.v_lim
        for spine in ax.spines.values():
            spine.set_visible(True)
            spine.set_linewidth(1)
        ax.set_xlim(h_lim)
        ax.set_ylim(v_lim)
        ax.grid(True)

        for i in self.components:
            if i in self.axes:
                if self.axes[i] == 0:
                    ax.set_xlabel(self.names[i], fontsize=font)
                if self.axes[i] == 1:
                    ax.set_ylabel(self.names[i], fontsize=font)

        if n_vec_h is None:
            n_vec_h = n_vec
            arrow_scale_h = n_vec_h
        if n_vec_v is None:
            n_vec_v = n_vec
            arrow_scale_v = n_vec_v

        if arrow_scale is None:
            arrow_scale = n_vec
        else:
            if arrow_scale_h is None:
                arrow_scale_h = arrow_scale
            if arrow_scale_v is None:
                arrow_scale_v = arrow_scale

        X = np.linspace(h_lim[0], h_lim[1], n_vec_h)
        Y = np.linspace(v_lim[0], v_lim[1], n_vec_v)

        xx, yy = np.meshgrid(X, Y)
        uu, vv = np.meshgrid(X, Y)

        for i in range(np.size(xx, 0)):
            for j in range(np.size(yy, 1)):
                uu[i, j], vv[i, j] = self.rate([xx[i, j], yy[i, j]])

        mag = np.hypot(uu, vv)
        un, vn = uu, vv
        with np.errstate(invalid='ignore'):
            un = np.where(np.isnan(un), np.nan,
                          np.where(mag > 0,
                                   np.divide(uu, mag), 0))
            vn = np.where(np.isnan(vn), np.nan,
                          np.where(mag > 0,
                                   np.divide(vv, mag), 0))
        ax.quiver(X, X, un, vn, color='#888888',
                  scale=arrow_scale,
                  alpha=0.4)
        return fig, ax, plt

    draw_rate_field = plot_rate_field


System = ReactiveSystem = ReactionSystem
